local addonName = ...
_G[addonName] = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")

local SpecStorage = _G[addonName]
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName)

-- GUI variables
local SpecStorageFrame
local selectedTreeChild = nil

-- Talent variables
local talentNameToIndex = {}
local talentIDToIndex = {}

-- Glyph variables
local liveMajorGlyphs = {}
local glyphList = {}
local glyphNameToIndex = {}

-- Constants
local numTalentTiers = MAX_TALENT_TIERS
local numTalentColumns = NUM_TALENT_COLUMNS
local numMajorGlyphs = 3

-- 1 = Primary, 2 = Secondary
local activeSpec = nil
-- Name of the active spec
local currentSpecName = nil

--[[
Frame Layout

SpecStorageFrame
	MainGroup
		Header
		TreeGroup
			ScrollFrame
				Saved Specs
					BossRows
				Add New Spec
					Edit boxes
					Dropdowns
					Add Button
				ButtonGroup
					Load Live Spec
					Clear Selection
]]--

-- Basic data broker support
local ldb = LibStub("LibDataBroker-1.1"):NewDataObject("SpecStorage", {
	type = "launcher",
	label = addonName,
	icon = "Interface\\icons\\spell_nature_enchantarmor",
	OnClick = function(self, button)
		if ( button == "LeftButton" ) then
			SpecStorage:Toggle()
		end
	end,
	OnTooltipShow = function(self)
		self:AddLine(addonName)
		self:AddLine("|cffffffcc" .. L["Left-click"] .. "|r " .. L["to toggle"])
	end,
})

-- Database defaults
local defaults = {
	class = {
	},
}

--------------------------
-- Core Addon Functions --
--------------------------

function SpecStorage:OnInitialize()
	-- Called when the addon is loaded
	self:RegisterChatCommand("specstorage", "ChatCommand")
	self:RegisterChatCommand("specs", "ChatCommand")
	self.db = LibStub("AceDB-3.0"):New("SpecStorageDB", defaults)
end

function SpecStorage:OnEnable()
	-- Called when the addon is enabled
	self:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
	self:SetupSpecVariable()
	self:SetupGlyphList()
	self:CreateMainFrame()
end

function SpecStorage:OnDisable()
	-- Called when the addon is disabled
	self:UnregisterAllEvents()
end

function SpecStorage:ChatCommand(msg)
	self:Toggle()
end

function SpecStorage:Toggle()
	if ( SpecStorageFrame:IsVisible() ) then
		SpecStorageFrame:Hide()
	else
		SpecStorageFrame:Show()
		self:RefreshDisplay()
	end
end

------------------------
-- Reusable Functions --
------------------------

function SpecStorage:TableIsEmpty(table)
	return next(table) == nil
end

function SpecStorage:PairsByKeys(t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
		table.sort(a, f)
		local i = 0
		local iter = function ()
			i = i + 1
			if a[i] == nil then return nil
				else return a[i], t[a[i]]
			end
		end
	return iter
end

function SpecStorage:SetupSpecVariable()
	local currentSpec = GetSpecialization()
	activeSpec = GetActiveSpecGroup()
	currentSpecName = currentSpec and select(2, GetSpecializationInfo(currentSpec))
end

function SpecStorage:SetupTooltip(item)
	local itemName = item:GetUserData("Name")
	GameTooltip:SetOwner(item.frame, "ANCHOR_TOPRIGHT")
	GameTooltip:SetText(itemName)
	GameTooltip:Show()
end

function SpecStorage:SetupTalentNameToIndex()
	if ( self:TableIsEmpty(talentNameToIndex) ) then
		for i = 1, numTalentTiers do
			for j = 1, numTalentColumns do
				local talentID, name = GetTalentInfo(i, j, activeSpec)
				talentIDToIndex[talentID] = j
				talentNameToIndex[name] = j
			end
		end
	end
end

function SpecStorage:SetupGlyphList()
	-- name, glyphType, isKnown, icon, glyphId, glyphLink, spec, specMatches = GetGlyphInfo(index)
	glyphList = {}
	glyphNameToIndex = {}
	local i = 1
	local name, glyphType, _, _, _, _, _, specMatches = GetGlyphInfo(i)

	-- Loop through glyphs until either hit minor glyphs or run out of glyphs
	while ( name ) do
		-- Ignore headers
		if ( name ~= "header" ) then
			if ( glyphType == GLYPH_TYPE_MINOR ) then
				break
			end
			if ( specMatches == true ) then
				table.insert(glyphList, name)
			end
		end
		i = i + 1
		name, glyphType, _, _, _, _, _, specMatches = GetGlyphInfo(i)
	end

	-- Glyphs will be out of order if there are some unknown ones
	table.sort(glyphList)

	-- Build the reverse lookup array
	for k, v in pairs(glyphList) do
		glyphNameToIndex[v] = k
	end

	table.insert(glyphList, "-" .. L["Clear Selection"] .. "-")
end

function SpecStorage:SetupLiveGlyphs()
	local glyphCount = 1

	for i = 1, NUM_GLYPH_SLOTS do
		local enabled, glyphType, _, _, _, glyphId = GetGlyphSocketInfo(i)
		if ( enabled and glyphType == GLYPH_TYPE_MAJOR ) then
			if ( glyphId ) then
				liveMajorGlyphs[glyphCount] = self:GetGlyphDetails(glyphId)
			else
				liveMajorGlyphs[glyphCount] = L["Blank"]
			end
			glyphCount = glyphCount + 1
		end
	end
end

function SpecStorage:GetNameAndTexture(type, id)
	if ( not id ) then
		return L["Blank"], "Interface\\icons\\INV_MISC_QUESTIONMARK"
	end

	if ( type == "Talent" ) then
		return select(2, GetTalentInfoByID(id))
	end

	if ( type == "Glyph" ) then
		return self:GetGlyphDetails(id)
	end
end

-- Returns glyphId if a name is passed or name and icon if an Id is passed
function SpecStorage:GetGlyphDetails(idOrName)
	local i = 1
	local name, glyphType, _, icon, glyphId = GetGlyphInfo(i)

	-- Loop through glyphs until either hit minor glyphs, run out of glyphs or the desired glyph is found
	while ( name ) do
		-- Ignore headers
		if ( name ~= "header" ) then
			if ( glyphType == GLYPH_TYPE_MINOR ) then
				break
			end

			if ( idOrName == name ) then
				return glyphId
			end

			if ( idOrName == glyphId ) then
				return name, icon
			end
		end

		i = i + 1
		name, glyphType, _, icon, glyphId = GetGlyphInfo(i)
	end

	return nil
end

---------------------
-- Event Functions --
---------------------

function SpecStorage:PLAYER_SPECIALIZATION_CHANGED()
	self:SetupSpecVariable()
	self:SetupGlyphList()

	if ( SpecStorageFrame:IsVisible() ) then
		self:RefreshDisplay()
	end
end

-------------------
-- GUI Functions --
-------------------

function SpecStorage:CreateMainFrame()
	local frame = AceGUI:Create("Window")
	frame:SetTitle("Spec Storage")
	frame:SetLayout("Fill")
	frame:SetWidth(775)
	frame.frame:SetFrameStrata("MEDIUM")
	SpecStorageFrame = frame
	SpecStorageFrame:Hide()
end

function SpecStorage:SetupUI()
	if ( not currentSpecName ) then
		self:Print(L["No specialization picked"])
		return
	end

	local mainGroup = AceGUI:Create("SimpleGroup")
	mainGroup:SetLayout("Flow")
	mainGroup:SetFullHeight(true)
	mainGroup:SetFullWidth(true)

	local header = AceGUI:Create("Heading")
	header:SetText(gsub(L["Saved specs for %s"], "%%s", currentSpecName))
	header:SetFullWidth(true)
	mainGroup:AddChild(header)

	local treeGroup = AceGUI:Create("TreeGroup")
	treeGroup:SetLayout("Fill")
	treeGroup:SetFullHeight(true)
	treeGroup:SetFullWidth(true)
	treeGroup:SetTree(self:SetupTree())
	treeGroup:SetCallback("OnGroupSelected", function(...) SpecStorage:SelectTreeChild(...) end)
	treeGroup:SelectByPath(1)

	mainGroup:AddChild(treeGroup)
	SpecStorageFrame:AddChild(mainGroup)
end

function SpecStorage:SetupTree()
	if ( not self.db.class[currentSpecName] ) then
		self.db.class[currentSpecName] = {}
	end

	local tree = {}
	local addNewGroup = {}
	addNewGroup.value = "Add"
	addNewGroup.text = L["Add New"]
	table.insert(tree, addNewGroup)

	for instance, bosses in self:PairsByKeys(self.db.class[currentSpecName]) do
		local instanceGroup = {}
		instanceGroup.children = {}
		for bossName, specDetails in self:PairsByKeys(bosses) do
			local boss = {}
			boss.value = bossName
			boss.text = bossName
			table.insert(instanceGroup.children, boss)
		end
		instanceGroup.value = instance
		instanceGroup.text = instance
		table.insert(tree, instanceGroup)
	end

	return tree
end

function SpecStorage:SelectTreeChild(parent, event, selection)
	local selectedParent, selectedChild = ("\001"):split(selection)
	parent:ReleaseChildren()

	local scrollFrame = AceGUI:Create("ScrollFrame")
	scrollFrame:SetLayout("List")
	scrollFrame:SetFullWidth(true)

	if ( selectedParent == "Add" or not selectedChild ) then
		selectedTreeChild = 1
	else
		selectedTreeChild = 2
		self:SetupSavedList(scrollFrame, selectedParent, selectedChild)
	end

	self:SetupAddNewRow(scrollFrame)
	self:SetupButtonGroup(scrollFrame)

	parent:AddChild(scrollFrame)
end

function SpecStorage:SetupIcon(name, texture)
	local iconSize = 40
	local iconFrameSize = 50
	local icon = AceGUI:Create("InteractiveLabel")

	icon:SetHeight(iconFrameSize)
	icon:SetWidth(iconFrameSize)
	icon:SetImage(texture)
	icon:SetImageSize(iconSize, iconSize)
	icon:SetUserData("Name", name)
	icon:SetCallback("OnEnter", function(self, _) SpecStorage:SetupTooltip(self) end)
	icon:SetCallback("OnLeave", function() GameTooltip:Hide() end)

	-- Mark glyph/talent as inactive until proven to be active
	icon.image:SetDesaturated(true)

	return icon
end

function SpecStorage:SetupSavedList(parent, instanceName, bossName)
	self:SetupLiveGlyphs()

	local container = AceGUI:Create("SimpleGroup")
	container:SetFullWidth(true)
	container:SetLayout("List")

	if not (self.db.class[currentSpecName]) then
		self.db.class[currentSpecName] = {}
	end

	local bossRow = AceGUI:Create("InlineGroup")
	bossRow:SetLayout("Flow")
	bossRow:SetTitle(instanceName..": "..bossName)
	bossRow:SetFullWidth(true)

	for i = 1, numTalentTiers do
		local talentID = self.db.class[currentSpecName][instanceName][bossName]["Tier"..i.."Talent"]
		local talentName, talentTexture = self:GetNameAndTexture("Talent", talentID)
		local talentIcon = self:SetupIcon(talentName, talentTexture)

		if ( talentName == L["Blank"] ) then
			-- Blanks are wildcards so always active
			talentIcon.image:SetDesaturated(false)
		else
			local unpicked, talentIndex = GetTalentRowSelectionInfo(i)
			if ( not unpicked ) then
				local name = select(2, GetTalentInfoByID(talentIndex))
				if (name == talentName) then
					-- Saved talent is active
					talentIcon.image:SetDesaturated(false)
				end
			end
		end
		bossRow:AddChild(talentIcon)
	end

	-- Gap between talents and glyphs
	local spacer = AceGUI:Create("Label")
	spacer:SetWidth(15)
	bossRow:AddChild(spacer)

	for i = 1, numMajorGlyphs do
		local glyphID = self.db.class[currentSpecName][instanceName][bossName]["MajorGlyph"..i]
		local glyphName, glyphTexture = self:GetNameAndTexture("Glyph", glyphID)
		local glyphIcon = self:SetupIcon(glyphName, glyphTexture)

		-- Loop through live major glyphs to see if saved glyph is active
		for j = 1, numMajorGlyphs do
			if ( glyphName == L["Blank"] ) then
				-- Blanks are wildcards so always active
				glyphIcon.image:SetDesaturated(false)
				break
			end
			if ( liveMajorGlyphs[j] == glyphName ) then
				-- Saved glyph is active
				glyphIcon.image:SetDesaturated(false)
				break
			end
		end
		bossRow:AddChild(glyphIcon)
	end

	local editButton = AceGUI:Create("Button")
	editButton:SetWidth(100)
	editButton:SetText(L["Edit"])
	editButton:SetCallback("OnClick", function() SpecStorage:EditingStoredSpec(instanceName, bossName) end)
	bossRow:AddChild(editButton)

	local deleteButton = AceGUI:Create("Button")
	deleteButton:SetWidth(100)
	deleteButton:SetText(L["Delete"])
	deleteButton:SetCallback("OnClick", function() SpecStorage:DeleteStoredSpec(instanceName, bossName) end)
	bossRow:AddChild(deleteButton)

	container:AddChild(bossRow)

	parent:AddChild(container)
end

function SpecStorage:SetupAddNewRow(parent)
	local container = AceGUI:Create("InlineGroup")
	container:SetLayout("Flow")
	container:SetFullWidth(true)
	container:SetTitle(L["Add/Edit a spec"])

	local instanceName = AceGUI:Create("EditBox")
	instanceName:SetLabel(L["Instance name"])
	instanceName:SetUserData("Instance", "Name")
	container:AddChild(instanceName)

	local bossName = AceGUI:Create("EditBox")
	bossName:SetLabel(L["Boss name"])
	bossName:SetUserData("Boss", "Name")
	container:AddChild(bossName)

	for i = 1, numTalentTiers do
		local talentDropDown = self:CreateTalentDropdown(i)
		talentDropDown:SetLabel(gsub(L["Tier %s Talent"], "%%s", i))
		talentDropDown:SetUserData("Tier", i)
		container:AddChild(talentDropDown)
	end

	for i = 1, numMajorGlyphs do
		local glyphDropDown = self:CreateGlyphDropdown()
		glyphDropDown:SetLabel(gsub(L["Major glyph %s"], "%%s", i))
		glyphDropDown:SetUserData("Glyph", i)
		container:AddChild(glyphDropDown)
	end

	local saveButton = AceGUI:Create("Button")
	saveButton:SetText(L["Save selection"])
	saveButton:SetCallback("OnClick", function() SpecStorage:AddSelection() end)

	container:AddChild(saveButton)

	parent:AddChild(container)
end

function SpecStorage:SetupButtonGroup(parent)
	local buttonGroup = AceGUI:Create("SimpleGroup")
	buttonGroup:SetLayout("Flow")
	buttonGroup:SetFullWidth(true)

	local loadLiveButton = AceGUI:Create("Button")
	loadLiveButton:SetText(L["Load Live Spec"])
	loadLiveButton:SetCallback("OnClick", function() SpecStorage:LoadLiveSpec() end)
	buttonGroup:AddChild(loadLiveButton)

	local clearButton = AceGUI:Create("Button")
	clearButton:SetText(L["Clear Talents and Glyphs"])
	clearButton:SetCallback("OnClick", function() SpecStorage:ClearAll() end)
	buttonGroup:AddChild(clearButton)

	parent:AddChild(buttonGroup)
end


function SpecStorage:GetTalentList(tier)
	-- talentID, name, texture, selected, available = GetTalentInfo(tier, column, talentGroup [, isInspect, inspectedUnit])
	local talentList = {}

	for i = 1, numTalentColumns do
		local name = select(2, GetTalentInfo(tier, i, activeSpec))
		table.insert(talentList, name)
	end

	table.insert(talentList, "-" .. L["Clear Selection"] .. "-")

	return talentList
end

function SpecStorage:CreateTalentDropdown(tier)
	local talentDropDown = AceGUI:Create("Dropdown")
	talentDropDown:SetWidth(150)
	talentDropDown:SetList(self:GetTalentList(tier))
	talentDropDown:SetCallback("OnValueChanged", function(self)
			-- Last in the dropdown is the clear selection option so clear the selection
			if ( self:GetValue() == numTalentColumns + 1 ) then
				self:SetValue(0)
			end
		end)

	return talentDropDown
end

function SpecStorage:CreateGlyphDropdown()
	local glyphDropDown = AceGUI:Create("Dropdown")
	glyphDropDown:SetWidth(150)
	glyphDropDown:SetList(glyphList)
	glyphDropDown:SetCallback("OnValueChanged", function(self)
			-- Last in the dropdown is the clear selection option so clear the selection
			if ( self:GetValue() == #glyphList ) then
				self:SetValue(0)
			end
		end)

	return glyphDropDown
end

function SpecStorage:RefreshDisplay()
	if ( #SpecStorageFrame.children > 0 ) then
		SpecStorageFrame:ReleaseChildren()
	end
	self:SetupUI()
end

----------------------
-- Button Functions --
----------------------

function SpecStorage:AddSelection()
	local instanceName
	local bossName
	local tierTalents = {}
	local glyphs = {}

	for objectIndex,objectAttributes in pairs(SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children) do
		for objectAttribute,objectAttributeTable in pairs(objectAttributes) do
			if ( objectAttribute == "userdata" ) then
				for key,value in pairs(objectAttributeTable) do
					if ( key == "Instance" ) then
						instanceName = SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:GetText()
					elseif ( key == "Boss" ) then
						bossName = SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:GetText()
					elseif ( key == "Tier" ) then
						local talentNumber = SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:GetValue()
						if ( talentNumber and talentNumber ~= 0 ) then
							local talentID = GetTalentInfo(value, talentNumber, activeSpec)
							table.insert(tierTalents, value, talentID)
						end
					elseif ( key == "Glyph" ) then
						local glyphNumber = SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:GetValue()
						if ( glyphNumber and glyphNumber ~= 0 ) then
							local glyphID = self:GetGlyphDetails(glyphList[glyphNumber])
							table.insert(glyphs, value, glyphID or L["Blank"]) -- fallback to a blank glyph if search fails
						end
					end
				end
			end
		end
	end

	-- Check all fields are filled
	if ( instanceName and instanceName ~= "" and bossName and bossName ~= "" ) then
		if ( not self.db.class[currentSpecName] ) then
			self.db.class[currentSpecName] = {}
		end
		if ( not self.db.class[currentSpecName][instanceName] ) then
			self.db.class[currentSpecName][instanceName] = {}
		end
		if ( not self.db.class[currentSpecName][instanceName][bossName] ) then
			self.db.class[currentSpecName][instanceName][bossName] = {}
		end

		for i = 1, numTalentTiers do
			self.db.class[currentSpecName][instanceName][bossName]["Tier" .. i .. "Talent"] = tierTalents[i]
		end

		for i = 1, numMajorGlyphs do
			self.db.class[currentSpecName][instanceName][bossName]["MajorGlyph" .. i] = glyphs[i]
		end

		self:RefreshDisplay()
		SpecStorageFrame.children[1].children[2]:SelectByValue(instanceName.."\001"..bossName)
	else
		self:Print(L["Please enter a instance and boss name"])
	end
end

function SpecStorage:EditingStoredSpec(instanceName, bossName)
	self:SetupTalentNameToIndex()

	for objectIndex,objectAttributes in pairs(SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children) do
		for objectAttribute,objectAttributeTable in pairs(objectAttributes) do
			if ( objectAttribute == "userdata" ) then
				for key,value in pairs(objectAttributeTable) do
					if ( key == "Instance" ) then
						SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetText(instanceName)
					elseif ( key == "Boss" ) then
						SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetText(bossName)
					elseif ( key == "Tier" ) then
						local talentID = self.db.class[currentSpecName][instanceName][bossName]["Tier" .. value .. "Talent"]
						local talentName = self:GetNameAndTexture("Talent", talentID)
						local column = talentNameToIndex[talentName] or 0
						SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(column)
					elseif ( key == "Glyph" ) then
						local glyphId = self.db.class[currentSpecName][instanceName][bossName]["MajorGlyph" .. value]
						local indexValue = 0
						if ( glyphId ~= L["Blank"] ) then
							local glyphName = self:GetGlyphDetails(glyphId)
							if ( glyphName ) then
								indexValue = glyphNameToIndex[glyphName]
							end
						end
						SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(indexValue)
					end
				end
			end
		end
	end
end

function SpecStorage:DeleteStoredSpec(instanceName, bossName)
	StaticPopupDialogs["Delete_Spec"] = {
		text = L["Are you sure you wish to delete?"],
		button1 = L["Yes"],
		button2 = L["No"],
		OnAccept = function()
			self.db.class[currentSpecName][instanceName][bossName] = nil
			if ( self:TableIsEmpty(self.db.class[currentSpecName][instanceName]) ) then
				-- All bosses for the instance are now gone
				self.db.class[currentSpecName][instanceName] = nil
			else
				SpecStorageFrame.children[1].children[2]:SelectByValue(instanceName)
			end
			self:RefreshDisplay()
		end,
		timeout = 10,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,
	}
	StaticPopup_Show("Delete_Spec")
end


function SpecStorage:LoadLiveSpec()
	self:SetupTalentNameToIndex()
	self:SetupLiveGlyphs()

	for objectIndex,objectAttributes in pairs(SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children) do
		for objectAttribute,objectAttributeTable in pairs(objectAttributes) do
			if ( objectAttribute == "userdata" ) then
				for key,value in pairs(objectAttributeTable) do
					if ( key == "Tier" ) then
						local unpicked, talentIndex = GetTalentRowSelectionInfo(value)
						if unpicked then
							SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(0)
						else
							SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(talentIDToIndex[talentIndex])
						end
					elseif ( key == "Glyph" ) then
						local glyphName = liveMajorGlyphs[tonumber(value)]
						if ( glyphName == L["Blank"] ) then
							SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(0)
						else
							local indexValue = glyphNameToIndex[glyphName]
							SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(indexValue)
						end
					end
				end
			end
		end
	end
end

function SpecStorage:ClearAll()
	for objectIndex,objectAttributes in pairs(SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children) do
		for objectAttribute,objectAttributeTable in pairs(objectAttributes) do
			if ( objectAttribute == "userdata" ) then
				for key,value in pairs(objectAttributeTable) do
					if ( key == "Tier" ) then
						SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(0)
					elseif ( key == "Glyph" ) then
						SpecStorageFrame.children[1].children[2].children[1].children[selectedTreeChild].children[objectIndex]:SetValue(0)
					end
				end
			end
		end
	end
end
